# Copyright 2025 Remy Blank <remy@c-space.org>
# SPDX-License-Identifier: MIT

import datetime
from docutils import nodes
from docutils.parsers.rst import directives
from sphinx.util import docutils, logging

from . import __version__, report_exceptions, UniqueChecker
from .. import util

_log = logging.getLogger(__name__)


def setup(app):
    app.add_directive('poll', Poll)
    app.add_node(poll, html=(visit_poll, depart_poll))
    app.add_node(answers, html=(visit_answers, depart_answers))
    app.add_node(answer, html=(visit_answer, depart_answer))
    app.add_env_collector(UniqueChecker('poll-id',
        lambda doctree: ((n, n['id']) for n in doctree.findall(poll)),
        "{poll}: Duplicate poll ID"))
    app.connect('html-page-context', add_js)
    return {
        'version': __version__,
        'parallel_read_safe': True,
        'parallel_write_safe': True,
    }


class Poll(docutils.SphinxDirective):
    required_arguments = 1
    option_spec = {
        'mode': lambda c: directives.choice(c, ('single', 'multi')),
        'number': lambda c: directives.choice(c,
            ('none', 'decimal', 'lower-alpha', 'upper-alpha')),
        'close-after': directives.unchanged,
        'class': directives.class_option,
    }
    has_content = True

    @report_exceptions
    def run(self):
        children = self.parse_content_to_nodes()
        if any(True for c in children for n in c.findall(poll)):
            raise Exception("{poll}: Must not contain {poll}")
        if not children or not isinstance(children[-1], nodes.bullet_list):
            raise Exception("{poll}: Must end with a bullet list")
        ans = answers('', *(answer('', *a) for a in children[-1]))
        # Find and remove ':' prefixes that tag solutions.
        for a in ans:
            for n in a.findall(nodes.Text):
                if n.startswith(':'):
                    a['solution'] = True
                    p = n.parent
                    if len(n) > 1:
                        p.replace(n, n.__class__(n[1:]))
                    else:
                        p.remove(n)
                        if not p.children: p.parent.remove(p)
                break
        node = poll('', *children[:-1], ans)
        self.set_source_info(node)
        node['id'] = self.arguments[0]
        node['mode'] = self.options.get('mode', 'single')
        if (v := self.options.get('close-after', '15m')) != 'never':
            node['close-after'] = util.parse_duration(v)
        node['number'] = self.options.get('number', 'upper-alpha')
        node['classes'] += self.options.get('class', [])
        return [node]


class poll(nodes.Body, nodes.Element): pass
class answers(nodes.Sequential, nodes.Element): pass
class answer(nodes.Part, nodes.Element): pass


def visit_poll(self, node):
    kwargs = {'data-id': node['id'], 'data-mode': node['mode']}
    if (v := node.get('close-after')) is not None:
        kwargs['data-close-after'] = v // datetime.timedelta(milliseconds=1)
    self.body.append(self.starttag(
        node, 'div', suffix='', classes=['tdoc-poll', f'num-{node['number']}'],
        **kwargs))


def depart_poll(self, node):
    self.body.append('</div>\n')


def visit_answers(self, node):
    cols = 5 if node.parent['number'] != 'none' else 4
    self.body.append(f"""
<table class="tdoc-poll-answers table">\
<thead><tr><th class="tdoc-poll-header" colspan="{cols}"><div>\
<div class="stats">\
<div class="voters" title="Voters">\
<span class="tdoc fa-user"></span><span></span></div>\
<div class="votes" title="Votes">\
<span class="tdoc fa-check-to-slot"></span><span></span></div>\
<div class="closed" title="The poll is closed">\
<span class="tdoc fa-lock"></span></div>\
</div><div class="controls">\
<button class="tdoc-open fa-play"></button>\
<button class="tdoc-results fa-eye"></button>\
<button class="tdoc-solutions fa-check"></button>\
<button class="tdoc-clear fa-trash"\
 title="Clear votes (Ctrl+click to clear all)"></button>\
</div></div></th></thead><tbody>\
""")


def depart_answers(self, node):
    self.body.append('</tbody></table>')


def visit_answer(self, node):
    num = '<td class="tdoc-poll-num"></td>' \
          if node.parent.parent['number'] != 'none' else ''
    s = ' s' if node.get('solution') else ''
    self.body.append(f"""\
<tr>{num}\
<td class="tdoc-poll-sel{s}"><span class="tdoc fa-circle-check"></span></td>\
<td class="tdoc-poll-ans">\
""")


def depart_answer(self, node):
    self.body.append("""\
</td><td class="tdoc-poll-cnt"></td><td class="tdoc-poll-pct"></td></tr>\
""")


def add_js(app, page, template, context, doctree):
    if doctree and any(doctree.findall(poll)):
        app.add_js_file('tdoc/poll.js', type='module')



# Copyright 2024 Remy Blank <remy@c-space.org>
# SPDX-License-Identifier: MIT

import ast
import asyncio
import contextlib
import contextvars
import inspect
import io
import sys
import traceback

import js
from pyodide.ffi import run_sync

try:
    from polyscript import xworker
except ImportError:
    xworker = None

run_id_var = contextvars.ContextVar('run_id', default=None)
run_id = run_id_var.get

tasks = {}
env = {'__name__': '__main__'}

if xworker is not None:
    js_input = xworker.sync.input
    js_render = xworker.sync.render
    js_write = xworker.sync.write

    def js_setup_canvas():
        raise Exception(
            "setup_canvas() is only supported in the 'main' environment")


def public(fn):
    """Make a function available in the client code environment."""
    env[fn.__name__] = fn
    return fn


def export(fn):
    """Export a function to JavaScript."""
    if xworker is not None: setattr(xworker.sync, fn.__name__, fn)
    return fn


def linenos(obj):
    """Return the start and past-the-end lines of an object."""
    lines, start = inspect.getsourcelines(obj)
    return start, start + len(lines)


class ContextProxy:
    """A proxy object that forwards attributes accesses to a context value."""
    __slots__ = ('_ContextProxy__var',)

    def __init__(self, var):
        super().__setattr__('_ContextProxy__var', var)

    def __getattr__(self, name):
        return getattr(self.__var.get(), name)

    def __setattr__(self, name, value):
        return setattr(self.__var.get(), name, value)

    def __delattr__(self, name):
        return delattr(self.__var.get(), name)

    def __dir__(self):
        return self.__var.get().__dir__()


class OutStream(io.RawIOBase):
    """An output stream that forwards writes to JavaScript."""
    def __init__(self, stream):
        self.stream = stream

    def writable(self): return True

    def write(self, data, /):
        size = len(data)
        if size > 0: js_write(run_id(), self.stream, data)
        return size


stdout_var = contextvars.ContextVar(
    'stdout',
    default=io.TextIOWrapper(io.BufferedWriter(OutStream(1)),
                             line_buffering=True))
stderr_var = contextvars.ContextVar(
    'stderr',
    default=io.TextIOWrapper(io.BufferedWriter(OutStream(2)),
                             line_buffering=True))
sys.stdout = ContextProxy(stdout_var)
sys.stderr = ContextProxy(stderr_var)


@public
@contextlib.contextmanager
def redirect(stdout=None, stderr=None):
    stdout_token, stderr_token = None, None
    if stdout is not None: stdout_token = stdout_var.set(stdout)
    if stderr is not None: stderr_token = stderr_var.set(stderr)
    try:
        yield
    finally:
        if stdout_token is not None: stdout_var.reset(stdout_token)
        if stderr_token is not None: stderr_var.reset(stderr_token)


_next_id = 0

@public
def new_id():
    """Generate a unique ID, usable in id= attributes."""
    global _next_id
    v = _next_id
    _next_id += 1
    return f'tdoc-id-{v}'


@public
def sleep(sec):
    """Sleep for the given number of seconds."""
    run_sync(asyncio.sleep(sec))


@public
def animation_time():
    """Return the current value of the animation timer."""
    return js.performance.now()


@public
def animation_frame():
    """Return a Promise that resolves at the next animation frame."""
    p = js.Promise.withResolvers()
    js.requestAnimationFrame(p.resolve)
    return p.promise


@public
def render(html, name=''):
    """Render some HTML in an output block."""
    if not isinstance(html, str): html = ''.join(html)
    return js_render(run_id(), html, name).then(lambda res: tuple(res))


@public
def setup_canvas():
    """Set up a <canvas> element for rendering."""
    return js_setup_canvas(run_id())


@public
async def input_line(prompt=None):
    """Request a single line of text from the user."""
    return await js_input(run_id(), 'line', prompt)


@public
async def input_text(prompt=None):
    """Request a multi-line text from the user."""
    return await js_input(run_id(), 'text', prompt)


@public
async def input_buttons(prompt, labels):
    """Present a list of buttons and wait for the user to click one of them."""
    return await js_input(run_id(), 'buttons', prompt, labels)


@public
async def pause(prompt=None, label="@icon{forward-step}"):
    """Present a button, and wait for the user to click it."""
    await js_input(run_id(), 'buttons-right', prompt, [label])


@public
def input(prompt=None):
    """Synchronously request a single line of text from the user."""
    return run_sync(input_line(prompt))


@export
async def run(run_id, blocks):
    """Run blocks of client code."""
    run_id_var.set(run_id)
    tasks[run_id] = asyncio.current_task()
    try:
        e = env.copy()
        for code, name in blocks:
            code = compile(code, name or '<unnamed>', 'exec',
                           flags=ast.PyCF_ALLOW_TOP_LEVEL_AWAIT)
            if (coro := eval(code, e)) is not None: await coro
    except BaseException as e:
        te = traceback.TracebackException.from_exception(e, compact=True)
        # Filter this function out of the stack trace
        for i, fs in enumerate(te.stack):
            if (fs.filename == __file__ and run_start <= fs.lineno < run_end
                    and fs.name == 'run'):
                del te.stack[i]
                break
        te.print()
    finally:
        # TODO: Allow registering and running cleanup handlers
        del tasks[run_id]


run_start, run_end = linenos(run)


@export
def stop(run_id):
    """Stop a running block of client code."""
    if (task := tasks.get(run_id)) is not None:
        task.cancel()



# Copyright 2024 Remy Blank <remy@c-space.org>
# SPDX-License-Identifier: MIT

import base64
import copy
import functools
import json
import pathlib

from docutils import nodes, statemachine
from myst_parser import mocking
from sphinx import config, locale
from sphinx.environment import collectors
from sphinx.util import docutils, fileutil, logging

from .. import __version__, deps

_log = logging.getLogger(__name__)
_messages = 'tdoc'
_ = locale.get_translation(_messages)

_base = pathlib.Path(__file__).parent.resolve().parent

_license_urls = {
    'CC0-1.0': 'https://creativecommons.org/publicdomain/zero/1.0/',
    'CC-BY-4.0': 'https://creativecommons.org/licenses/by/4.0/',
    'CC-BY-SA-4.0': 'https://creativecommons.org/licenses/by-sa/4.0/',
    'CC-BY-NC-4.0': 'https://creativecommons.org/licenses/by-nc/4.0/',
    'CC-BY-NC-SA-4.0': 'https://creativecommons.org/licenses/by-nc-sa/4.0/',
    'CC-BY-ND-4.0': 'https://creativecommons.org/licenses/by-nd/4.0/',
    'MIT': 'https://opensource.org/license/mit',
}

# BUG(myst_parser): MockState.parse_directive_block() [myst_parser] returns the
# content as a StringList, whereas Body.parse_directive_block() [docutils]
# returns a list. The StringList is constructed with source=content.source,
# which is a bound method and clearly wrong. Patch the method to unwrap the
# list.
def _parse_directive_block(self, *args, **kwargs):
    arguments, options, content, content_offset = \
        MockState_parse_directive_block(self, *args, **kwargs)
    if isinstance(content, statemachine.StringList): content = content.data
    return arguments, options, content, content_offset

MockState_parse_directive_block = mocking.MockState.parse_directive_block
mocking.MockState.parse_directive_block = _parse_directive_block


def to_base64(s):
    return base64.b64encode(s.encode('utf-8')).decode('utf-8').rstrip('=')


def names_option(arg):
    if arg is None: raise ValueError('no argument provided')
    return [nodes.fully_normalize_name(n) for n in arg.split()]


def report_exceptions(fn):
    @functools.wraps(fn)
    def wrapper(self, /, *args, **kwargs):
        try:
            return fn(self, *args, **kwargs)
        except Exception as e:
            if isinstance(self, Role):
                err = self.inliner.document.reporter.error(e, line=self.lineno)
                return [], [err]
            return [self.state.document.reporter.error(e, line=self.lineno)]
    return wrapper


def format_attrs(translator, /, **kwargs):
    return ' '.join(f'{k.replace('_', '-')}="{translator.attval(v)}"'
                    for k, v in sorted(kwargs.items()) if v is not None)


def format_data_attrs(translator, /, **kwargs):
    return ' '.join(f'data-tdoc-{k.replace('_', '-')}="{translator.attval(v)}"'
                    for k, v in sorted(kwargs.items()) if v is not None)


def setup(app):
    app.set_html_assets_policy('always')  # Ensure MathJax is always available
    app.add_event('tdoc-html-page-config')

    app.add_config_value('license', '', 'html', str)
    app.add_config_value(
        'license_url', lambda c: _license_urls.get(c.license, ''), 'html', str)
    app.add_config_value('tdoc', {}, 'html', dict)
    app.add_config_value('tdoc_api', '', 'html', str)
    app.add_config_value('tdoc_domain_storage', {}, 'html', dict)
    app.add_config_value('tdoc_enable_sab', 'no', 'html',
                         config.ENUM('no', 'cross-origin-isolation', 'sabayon'))
    app.add_config_value('tdoc_versions', {}, 'html', dict)

    app.add_html_theme('t-doc', str(_base))
    app.add_message_catalog(_messages, str(_base / 'locale'))

    app.connect('config-inited', on_config_inited)
    app.connect('builder-inited', on_builder_inited)
    app.connect('html-page-context', on_html_page_context, priority=499.9)
    app.connect('html-page-context', restore_mathjax, priority=500.1)
    if 'tdoc-dev' in app.tags:
        app.connect('html-page-context', add_terminate_button, priority=500.4)
    app.connect('html-page-context', add_draw_button, priority=500.6)
    app.connect('html-page-context', add_user_button, priority=500.7)
    app.connect('write-started', write_static_files)

    return {
        'version': __version__,
        'parallel_read_safe': True,
        'parallel_write_safe': True,
    }


def on_config_inited(app, config):
    cv = config.values['html_title']
    super(cv.__class__, cv).__setattr__('default', lambda c: c.project)
    config.templates_path.append(str(_base / 'templates'))

    # Add our own static paths, and a default one if it exists.
    config.html_static_path.append(str(_base / 'static'))
    config.html_static_path.append(str(_base / 'static.gen'))
    if '_static' not in config.html_static_path \
            and (app.confdir / '_static').exists():
        config.html_static_path.append('_static')

    # Override defaults in html_theme_options.
    opts = config.html_theme_options
    opts.setdefault('use_sidenotes', True)
    opts.setdefault('path_to_docs', 'docs')
    if opts.get('repository_url'):
        opts.setdefault('use_repository_button', True)
        opts.setdefault('use_source_button', True)
    opts.setdefault('use_download_button', False)


def on_builder_inited(app):
    # The config is used in domain.html.jinja.
    tdoc = tdoc_config(app, None, None)
    app.config.html_context['tdoc'] = json.dumps(tdoc, separators=(',', ':'))


def on_html_page_context(app, page, template, context, doctree):
    context['tdoc_version'] = __version__
    if v := app.config.license: context['license'] = v
    if v := app.config.license_url: context['license_url'] = v

    # Set up early and on-load JavaScript. Temporarily override mathjax_path
    # as per version specification, then restore it after
    # mathjax.install_mathjax() has run.
    tdoc = tdoc_config(app, page, doctree)
    mathjax = tdoc['versions'].pop('mathjax')
    if mathjax.startswith('/'): mathjax = f'../{mathjax[1:]}'
    context['tdoc_mathjax_path'] = app.config.mathjax_path
    app.config.mathjax_path = f'{mathjax}/tex-chtml-full.js'
    tdoc = json.dumps(tdoc, separators=(',', ':'))
    app.add_js_file(None, priority=0, body=f'const tdoc = {tdoc};')
    app.add_js_file('tdoc/early.js', priority=1)
    app.add_js_file('tdoc/load.js', type='module')


def restore_mathjax(app, page, template, context, doctree):
    app.config.mathjax_path = context['tdoc_mathjax_path']
    del context['tdoc_mathjax_path']


def tdoc_config(app, page, doctree):
    tdoc = {
        'conf': copy.deepcopy(app.config.tdoc),
        'domain_storage': copy.deepcopy(app.config.tdoc_domain_storage),
        'enable_sab': app.config.tdoc_enable_sab,
        'html_data': {},
    }
    if is_dev := 'tdoc-dev' in app.tags: tdoc['dev'] = True
    versions = tdoc['versions'] = \
        (app.env.metadata[page].get('versions') or {}).copy()
    for name, v in app.config.tdoc_versions.items():
        versions.setdefault(name, v)
    for name, info in deps.info.items():
        if '://' not in (v := versions.setdefault(name, info['version'])):
            versions[name] = f'/_cache/{name}/{v}' if is_dev else info['url'](v)
    if v := app.config.tdoc_api: tdoc['api_url'] = v
    app.emit('tdoc-html-page-config', page, tdoc, doctree)
    return tdoc


def add_terminate_button(app, page, template, context, doctree):
    if doctree is None: return
    context["header_buttons"].append({
        'type': 'javascript',
        'javascript': 'tdoc.terminateServer()',
        'tooltip': _("Terminate the local server"),
        'label': 'terminate',
    })


def add_draw_button(app, page, template, context, doctree):
    if doctree is None: return
    context["header_buttons"].append({
        'type': 'javascript',
        'javascript': 'tdoc.draw()',
        'tooltip': _("Draw"),
        'label': 'draw',
    })


def add_user_button(app, page, template, context, doctree):
    if doctree is None: return
    context["header_buttons"].append({
        'type': 'group',
        'label': 'user',
        'buttons': [{
            'type': 'link',
            'text': "Not signed in",
            'icon': 'fa fa-user',
            'label': 'user',
        }, {
            'type': 'javascript',
            'javascript': 'tdoc.login()',
            'text': "Sign in",
            'icon': 'fa fa-right-to-bracket',
            'label': 'login',
        }, {
            'type': 'javascript',
            'javascript': 'tdoc.logout()',
            'text': "Sign out",
            'icon': 'fa fa-right-from-bracket',
            'label': 'logout',
        }],
    })


def write_static_files(app, builder):
    if builder.format != 'html': return

    # The file must be at the root of the website, to avoid limiting the scope
    # of the service worker to _static.
    fileutil.copy_asset_file(_base / 'scripts' / 'tdoc-worker.js',
                             builder.outdir, force=True)


class Role(docutils.SphinxRole):
    def __new__(cls, *args, **kwargs):
        self = super().__new__(cls)
        return self(*args, **kwargs)


class UniqueChecker(collectors.EnvironmentCollector):
    def __init__(self, name, iter_nodes, err):
        self.name = name
        self.iter_nodes = iter_nodes
        self.err = err

    def __repr__(self): return f'UniqueChecker({self.name})'
    def __call__(self): return self

    def enable(self, app):
        def init(app):
            if not hasattr(app.env, 'tdoc_unique'): app.env.tdoc_unique = {}
            # ID => (docname, location)
            app.env.tdoc_unique.setdefault(self.name, {})
        self.init_listener = app.connect('builder-inited', init)
        super().enable(app)

    def disable(self, app):
        super().disable(app)
        app.disconnect(self.init_listener)

    def clear_doc(self, app, env, docname):
        ids = env.tdoc_unique[self.name]
        for v in list(ids):
            if ids[v][0] == docname: del ids[v]

    def merge_other(self, app, env, docnames, other):
        ids = env.tdoc_unique[self.name]
        for v, (dn, loc) in other.tdoc_unique.get(self.name, {}).items():
            if dn not in docnames: continue
            if v not in ids:
                ids[v] = (dn, loc)
            else:
                _log.error(f"{self.err}: {v}", location=loc)

    def process_doc(self, app, doctree):
        ids = app.env.tdoc_unique[self.name]
        for node, v in self.iter_nodes(doctree):
            if not v: continue
            if v not in ids:
                ids[v] = (app.env.docname, (node.source, node.line))
            else:
                doctree.reporter.error(f"{self.err}: {v}", base_node=node)
